sessions = [{
    "1": {
        "type": "session",
        "source": {"id": "scope"},
        "id": "1",
        'profile': {"id": "1"}
    }
}]

profiles = [
    {"1": {'id': "1", "traits": {}}},
    {"2": {'id': "2", "traits": {}}},
]


class MockStorageCrud:

    def __init__(self, index, domain_class_ref, entity):
        self.index = index
        self.domain_class_ref = domain_class_ref
        self.entity = entity
        if index == 'session':
            self.data = sessions
        elif index == 'profile':
            self.data = profiles

    async def load(self):

        for item in self.data:
            if self.entity.id in item:
                return self.domain_class_ref(**item[self.entity.id])
        return None

    async def save(self):
        self.data.append({self.entity.id: self.entity.dict(exclude_unset=True)})

    async def delete(self):
        del(self.data[self.entity.id])


class EntityStorageCrud:

    def __init__(self, index, entity):
        self.index = index
        self.entity = entity
        if index == 'session':
            self.data = sessions
        elif index == 'profile':
            self.data = profiles

    async def load(self, domain_class_ref):

        for item in self.data:
            if self.entity.id in item:
                return domain_class_ref(**item[self.entity.id])
        return None

    async def save(self):
        self.data.append({self.entity.id: self.entity.dict(exclude_unset=True)})

    async def delete(self):
        del(self.data[self.entity.id])




from tracardi.domain.time_range_query import DatetimeRangePayload

# g = DatetimeRangePayload()
# print(g.get_dates())
# print()

a = {"minDate": {"absolute": None, "delta": {"type": "minus", "value": -2, "entity": "day"}, "now": None},
     "maxDate": {"absolute": None, "delta": {"type": "plus", "value": 1, "entity": "week"}, "now": None},
     "where": "undefined", "limit": 10, "timeZone": "Europe/Warsaw", "rand": "0.9398943914399749"}
q = DatetimeRangePayload(**a)
print(q.minDate.absolute)
d1, d2 = q.get_dates()
print(d1, d2)
# exit()
# a = {
#     "minDate": {
#         "absolute": {
#             "year": 2010,
#             "month": 1,
#             "date": 1,
#             "hour": 2,
#             "minute": 3,
#             "second": 4,
#             "meridiem": "PM"
#         },
#         "delta": {
#             "entity": "hour",
#             "value": -1,
#         }
#     },
#     "maxDate": {
#         "absolute": {
#             "year": 2020,
#             "month": 1,
#             "date": 1,
#             "hour": 2,
#             "minute": 3,
#             "second": 4,
#             "meridiem": "PM"
#         },
#     }
# }
#
# q = DatetimeRangePayload(**a)
# dd = q.get_dates()
# pprint(dd)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tracardi',
    version='0.6.3',
    description='Trackardi Customer Data Platform backend',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Risto Kowaczewski',
    author_email='risto.kowaczewski@gmail.com',
    packages=['tracardi'],
    install_requires=[
        'pydantic==1.8.2',
        'aiohttp==3.7.4.post0',
        'aiohttp[speedups]',
        'elasticsearch==7.13.4',
        'prodict==0.8.18',
        'tzlocal==2.1',
        'python-multipart==0.0.5',
        'lark==0.11.3',
        'dateparser==1.0.0',
        'dotty-dict==1.3.0',
        'pytz==2021.1',
        'tracardi-plugin-sdk>=0.6.2',
        'tracardi_graph_runner>=0.6.2',
        'device_detector==0.10',
        'tracardi-dot-notation',
        'pip>=21.2.4',
        'deepdiff>=5.5.0',

        'tracardi-key-counter',
        'tracardi-rabbitmq-publisher',
        'tracardi-weather',
        'tracardi-mongodb-connector',
        'tracardi-maxmind-geolite2',
        'tracardi-day-night-split',
        'tracardi-discord-webhook',
        'tracardi-remote-call',
        'tracardi-zapier-webhook'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=['tracardi'],
    include_package_data=True,
    python_requires=">=3.8",
)



# import asyncio
# from uuid import uuid4
# import tracardi.domain.entity as entity
# from tracardi.domain.profile import Profile
# from tracardi.domain.profile_traits import ProfileTraits
# from tracardi.domain.response import Response, ResponseItem
# from tracardi.event_server.service.persistence_service import PersistenceService
# from tracardi.service.storage.elastic_storage import ElasticStorage
# from tracardi.service.storage.index import resources
#
#
# async def response_data():
#     profile_id = str(uuid4())
#
#     profile = Profile(id=profile_id, traits=ProfileTraits(public={"a": 1}))
#     await profile.storage().save()
#
#     pr = entity.Entity(id=profile_id)
#     print(await pr.storage('profile').load(Profile))
#
#     response_id = str(uuid4())
#     response_entity = Response(id=response_id,
#                                data=[ResponseItem(index='profile', slices=['traits.public'])])
#
#     await response_entity.storage().save()
#
#     response_entity = entity.Entity(id=response_id)
#
#     response = await response_entity.storage('response').load(Response)  # type: Response
#     print(response)
#     for response_item in response.data:
#         index = response_item.index
#         slices = response_item.slices
#         if index in resources and resources[index].rel is not None:
#             storage = PersistenceService(ElasticStorage(index_key=index))
#             result = await storage.load_by(resources[index].rel, profile.id)
#             print(result)
#
#
# asyncio.run(response_data())



import asyncio
from pprint import pprint

from tracardi.domain.index import Index
from tracardi.domain.time_range_query import DatetimeRangePayload


async def main():
    a = {
        "minDate": {
            "delta": {
                "entity": "month",
                "value": -3,
            }
        }
    }
    q = DatetimeRangePayload(**a)
    es = Index("session")
    r = await es.histogram(q)
    pprint(r)

loop = asyncio.get_event_loop()
loop.run_until_complete(main())
loop.close()


from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi_dot_notation.dot_accessor import DotAccessor


class CutOutTraitAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'trait' not in kwargs:
            raise ValueError("Please define trait in config section.")

        if kwargs['trait'] == 'undefined':
            raise ValueError("Please define trait in config section. It has default value of undefined.")

        self.property = kwargs['trait']

    async def run(self, payload: dict):

        dot = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        return Result(port="trait", value=dot[self.property])


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.cut_out_trait_action',
            className='CutOutTraitAction',
            inputs=['payload'],
            outputs=["trait"],
            init={
                "trait": "undefined"
            },
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Cut out trait',
            desc='Returns defined property from payload.',
            type='flowNode',
            width=200,
            height=100,
            icon='property',
            group=["Traits"]
        )
    )



import base64
import json
from typing import List
from typing import Optional

from tracardi.domain.storage_result import StorageResult
from tracardi.service.storage.collection_crud import CollectionCrud
from tracardi.domain.named_entity import NamedEntity
from tracardi.service.storage.crud import StorageCrud


class Credential(NamedEntity):
    description: Optional[str] = ''
    type: str
    group: List[str] = ["General"]
    enabled: bool = True
    data: dict


class CredentialRecord(NamedEntity):
    description: Optional[str] = ''
    type: str
    enabled: bool = True
    data: str
    group: List[str] = ["General"]

    @staticmethod
    def encode(credential: Credential) -> 'CredentialRecord':
        json_data = json.dumps(credential.data)
        b64_data = base64.b64encode(json_data.encode('utf-8'))

        return CredentialRecord(
            id=credential.id,
            name=credential.name,
            description=credential.description,
            type=credential.type,
            enabled=credential.enabled,
            group=credential.group,
            data=b64_data
        )

    def decode(self) -> Credential:
        decoded = base64.b64decode(self.data)
        data = json.loads(decoded)

        return Credential(
            id=self.id,
            name=self.name,
            description=self.description,
            type=self.type,
            enabled=self.enabled,
            group=self.group,
            data=data
        )

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("credential", CredentialRecord, entity=self)


class CredentialRecords(list):

    def bulk(self) -> CollectionCrud:
        return CollectionCrud("credential", self)

    @staticmethod
    def decode(records: StorageResult) -> List[Credential]:
        return [CredentialRecord.decode(CredentialRecord(**record)) for record in records]



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner


class PayloadPlugin(ActionRunner):

    def __init__(self):
        pass

    async def run(self, payload):
        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        spec=Spec(
            module='task_plugins.payload.payload_plugin',
            className='PayloadPlugin',
            inputs=[],
            outputs=["payload"]
        ),
        metadata=MetaData(
            name='Start',
            desc='Loads payload into flow.',
            type='flowNode',
            width=100,
            height=100,
            icon='payload'
        )
    )



class StorageResult:
    def __init__(self, result=None):
        if result is None:
            self.total = 0
            self._hits = []
            self.chunk = 0
        else:
            self.total = result['hits']['total']['value']
            self._hits = result['hits']['hits']
            self.chunk = len(self._hits)

    def __repr__(self):
        return "hits {}, total: {}".format(self._hits, self.total)

    def __iter__(self):
        for hit in self._hits:
            row = hit['_source']
            row['id'] = hit['_id']
            yield row

    def dict(self):
        return {
            "total": self.total,
            "result": list(self)
        }

    def __len__(self):
        return self.chunk

# x = {
#     "took": 0,
#     "timed_out": False,
#     "_shards": {
#         "total": 1,
#         "successful": 1,
#         "skipped": 0,
#         "failed": 0
#     },
#     "hits": {
#         "total": {
#             "value": 1,
#             "relation": "eq"
#         },
#         "max_score": 0.2876821,
#         "hits": [
#             {
#                 "_index": "test1-tracardi-rule",
#                 "_type": "_doc",
#                 "_id": "my-rule",
#                 "_score": 0.2876821,
#                 "_source": {
#                     "id": "my-rule",
#                     "source": {
#                         "id": "mobile-app"
#                     },
#                     "trigger": "event.type==\"xxx1\"",
#                     "actions": [
#                         {
#                             "module": "tracardi.process_engine.actions.copy_all_action",
#                             "className": "CopyAllAction",
#                             "config": {
#                                 "runOnce": True
#                             }
#                         }
#                     ],
#                     "is_cyclic": False
#                 }
#             }
#         ]
#     }
# }
#
# a = StorageResult(x)
# for q in a:
#     print(q)



from tracardi_dot_notation.dot_accessor import DotAccessor
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner

from tracardi.process_engine.tql.condition import Condition


class IfAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'condition' not in kwargs:
            raise ValueError("Condition is not set. Define it in config section.")
        self.condition = kwargs['condition']

    async def run(self, payload: dict):

        if self.condition is None:
            raise ValueError("Condition is not set. Define it in config section.")

        dot = DotAccessor(self.profile, self.session, payload, self.event, self.flow)

        if Condition.evaluate(self.condition, dot):
            return Result(port="TRUE", value=payload), Result(port="FALSE", value=None)
        else:
            return Result(port="FALSE", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.if_action',
            className='IfAction',
            inputs=["payload"],
            outputs=["TRUE", "FALSE"],
            init={"condition": None},
            manual="if_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='If',
            desc='This a conditional action that conditionally runs a branch of workflow.',
            keywords=['condition'],
            type='flowNode',
            width=200,
            height=100,
            icon='if',
            editor='text',
            group=['Conditions']
        )
    )



from dotty_dict import dotty
from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result
from tracardi_dot_notation.dot_accessor import DotAccessor


class ReshapePayloadAction(ActionRunner):

    def __init__(self, **kwargs):
        self.mapping = kwargs

    async def run(self, payload):

        source = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        config = dotty(self.mapping)

        destination = dotty()
        for key, value in config.items():
            destination[key] = source[value]

        return Result(port="payload", value=destination.to_dict())


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.reshape_payload_action',
            className='ReshapePayloadAction',
            inputs=["payload"],
            outputs=['payload'],
            init={},
            manual="reshape_payload_action",
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Reshape payload',
            desc='Creates new payload from provided data. Configuration defines where the data should be copied.',
            type='flowNode',
            width=200,
            height=100,
            icon='copy-property',
            group=["Traits"]
        )
    )




from uuid import uuid4
from pydantic import BaseModel
import tracardi.service.storage.crud as crud


class Entity(BaseModel):
    id: str

    def storage(self, index) -> 'crud.EntityStorageCrud':
        return crud.EntityStorageCrud(index, entity=self)

    @staticmethod
    def new() -> 'Entity':
        return Entity(id=str(uuid4()))



from tracardi.domain.pii import PII
from tracardi.domain.profile import Profile, Profiles
from tracardi.domain.profile_stats import ProfileStats


def test_profile_merge():
    p1 = Profile(
        id="1",
        stats=ProfileStats(views=1, visits=2),
        traits={
            "private": {
                "a": 1,
                "b": 2
            },
            "public": {
                "a": 1,
                "b": 2
            }
        },
        pii=PII(
            name="john",
            surname="doe"
        ),
        segments=['segment-1'],
        consents={"all": "granted"}
    )

    p2 = Profile(
        id="2",
        stats=ProfileStats(views=2, visits=4),
        traits={
            "private": {
                "a": 2,
                "c": 1
            }
        },
        pii=PII(
            name="jonathan",
            surname="doe"
        )
    )

    p3 = Profile(
        id="3",
        traits={
            "private": {
                "a": [3],
                "c": 1
            },
            "public": {
                "a": 1,
                "b": 3
            }
        },
        segments=['segment-2'],
        consents={"all": "not-granted"}
    )

    profiles = [p1, p2]
    p = Profiles.merge(profiles, p3)

    assert p.consents == {'all': 'not-granted'}
    assert p.traits.private == {'b': 2, 'a': [1, 2, 3], 'c': 1}
    assert p.traits.public == {'b': [2, 3], 'a': 1}
    assert set(p.pii.name).intersection({'john', 'jonathan'}) == {'john', 'jonathan'}
    assert p.pii.surname == 'doe'
    assert p.mergedWith is None
    assert p.stats.views == 3
    assert p.stats.visits == 6



import os


class TracardiConfig:
    def __init__(self, env):
        self.track_debug = env['TRACK_DEBUG'] if 'TRACK_DEBUG' in env else False


class MemoryCacheConfig:
    def __init__(self, env):
        self.source_ttl = env['SOURCE_TTL'] if 'SOURCE_TTL' in env else 60


class ElasticConfig:

    def __init__(self, env):

        if 'ELASTIC_HOST' in env and not isinstance(env['ELASTIC_HOST'], str):
            raise ValueError("Env ELASTIC_HOST must be sting")

        if 'ELASTIC_HOST' in env and isinstance(env['ELASTIC_HOST'], str) and env['ELASTIC_HOST'].isnumeric():
            raise ValueError("Env ELASTIC_HOST must be sting")

        self.host = env['ELASTIC_HOST'] if 'ELASTIC_HOST' in env else '127.0.0.1'
        self.host = self.host.split(",")
        self.sniff_on_start = env['ELASTIC_SNIFF_ON_START'] if 'ELASTIC_SNIFF_ON_START' in env else None
        self.sniff_on_connection_fail = env[
            'ELASTIC_SNIFF_ON_CONNECTION_FAIL'] if 'ELASTIC_SNIFF_ON_CONNECTION_FAIL' in env else None
        self.sniffer_timeout = env['ELASTIC_SNIFFER_TIMEOUT'] if 'ELASTIC_SNIFFER_TIMEOUT' in env else None
        self.http_auth_username = env['ELASTIC_HTTP_AUTH_USERNAME'] if 'ELASTIC_HTTP_AUTH_USERNAME' in env else None
        self.http_auth_password = env['ELASTIC_HTTP_AUTH_PASSWORD'] if 'ELASTIC_HTTP_AUTH_PASSWORD' in env else None
        self.scheme = env['ELASTIC_SCHEME'] if 'ELASTIC_SCHEME' in env else None
        self.cafile = env['ELASTIC_CAFILE'] if 'ELASTIC_CAFILE' in env else None
        self.api_key = ('id', env['ELASTIC_API_KEY']) if 'ELASTIC_API_KEY' in env else None
        self.cloud_id = env['ELASTIC_CLOUD_ID'] if 'ELASTIC_CLOUD_ID' in env else None
        self.maxsize = env['ELASTIC_MAX_CONN'] if 'ELASTIC_MAX_CONN' in env else None
        self.http_compress = env['ELASTIC_HTTP_COMPRESS'] if 'ELASTIC_HTTP_COMPRESS' in env else None

        self.sql_translate_url = env[
            'ELASTIC_SQL_TRANSLATE_URL'] if 'ELASTIC_SQL_TRANSLATE_URL' in env else "/_sql/translate"
        self.sql_translate_method = env[
            'ELASTIC_SQL_TRANSLATE_METHOD'] if 'ELASTIC_SQL_TRANSLATE_METHOD' in env else "POST"


elastic = ElasticConfig(os.environ)
memory_cache = MemoryCacheConfig(os.environ)
tracardi = TracardiConfig(os.environ)



from typing import List, Optional

from tracardi.domain.named_entity import NamedEntity


class FlowMetaData(NamedEntity):
    description: str
    enabled: bool = True
    projects: Optional[List[str]] = ["General"]



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class MergeProfilesAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        if 'mergeBy' not in kwargs:
            raise ValueError("Field mergeBy is not set. Define it in config section.")
        self.merge_key = kwargs['mergeBy']
        self.merge_key = [key.lower() for key in self.merge_key]

        if len(self.merge_key) == 0:
            raise ValueError("Field mergeBy is empty and has no effect on merging. "
                             "Add merging key or remove this action from flow.")

        for key in self.merge_key:
            if not key.startswith('profile@'):
                raise ValueError(
                    f"Field mergeBy must define profile fields. Dot notation `{key}` does not start with profile@...")

    async def run(self, void):
        self.profile.operation.merge = self.merge_key
        return Result(value={}, port="void")


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.operations.merge_profiles_action',
            className='MergeProfilesAction',
            inputs=["void"],
            outputs=["void"],
            init={"mergeBy": []},
            manual="merge_profiles_action"
        ),
        metadata=MetaData(
            name='Merge profiles',
            desc='Merges profile in storage when flow ends. This operation is expensive so use it with caution, '
                 'only where this is new PII information added.',
            type='flowNode',
            width=200,
            height=100,
            icon='merge',
            group=["Operations"]
        )
    )



from time import time
from typing import Any

from pydantic import BaseModel

from tracardi.exceptions.exception import ExpiredException


class CacheItem(BaseModel):
    data: Any
    ttl: int = 60

    def __init__(self, **data: Any):
        if 'ttl' in data:
            data['ttl'] = time() + data['ttl']
        super().__init__(**data)

    def expired(self):
        return time() > self.ttl


class MemoryCache:

    def __init__(self):
        self.memory_buffer = {}

    def __contains__(self, key: str):
        if key in self.memory_buffer:
            if self.memory_buffer[key].expired():
                del self.memory_buffer[key]

        return key in self.memory_buffer

    def __getitem__(self, item: str) -> [CacheItem, None]:
        if item in self.memory_buffer:
            cache_item = self.memory_buffer[item]  # type: CacheItem
            if cache_item.expired():
                del self.memory_buffer[item]
                raise ExpiredException("MemoryCache item expired")
            return cache_item
        return None

    def __setitem__(self, key: str, value: CacheItem):
        if not isinstance(value, CacheItem):
            raise ValueError("MemoryCache item must be CacheItem type.")
        self.memory_buffer[key] = value



from datetime import datetime
from typing import Optional, Any

from pydantic import BaseModel


class Time(BaseModel):
    insert: Optional[datetime]

    def __init__(self, **data: Any):
        data['insert'] = datetime.utcnow()
        super().__init__(**data)




from typing import Optional

from tracardi.domain.named_entity import NamedEntity
from tracardi.service.storage.crud import StorageCrud


class Segment(NamedEntity):
    description: Optional[str] = ""
    eventType: Optional[str] = None
    condition: str
    enabled: bool = True

    def get_id(self) -> str:
        return self.name.replace(" ", "-").lower()

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("segment", Segment, entity=self)



import logging
from tracardi.event_server.service.persistence_service import PersistenceService
from tracardi.service.storage.elastic_storage import ElasticStorage


rules_logger = logging.getLogger('Rules')


class Rules(list):

    @staticmethod
    def storage() -> PersistenceService:
        return PersistenceService(ElasticStorage(index_key="rule"))



from datetime import datetime
from typing import Optional, Any
from tracardi_plugin_sdk.domain.register import Plugin

from tracardi.service.storage.crud import StorageCrud
from .entity import Entity
from .metadata import Metadata
from .settings import Settings
from .time import Time


class FlowActionPlugin(Entity):
    metadata: Optional[Metadata]
    plugin: Plugin
    settings: Optional[Settings] = Settings()

    def __init__(self, **data: Any):
        data['metadata'] = Metadata(
            time=Time(
                insert=datetime.utcnow()
            ))
        super().__init__(**data)

    # Persistence

    def storage(self) -> StorageCrud:
        return StorageCrud("action", FlowActionPlugin, entity=self)



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner


class JoinAction(ActionRunner):

    def __init__(self, *args, **kwargs):
        pass

    async def run(self, payload):
        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        debug=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.join_action',
            className='JoinAction',
            inputs=["payload"],
            outputs=["payload"],
            init=None,
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Join',
            desc='This action waits for previous actions to finish working. It passes payload '
                 'it recevied without any modifications.',
            type='flowNode',
            width=100,
            height=100,
            icon='join',
            group=["Flow"]
        )
    )



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.domain.result import Result
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi.domain.profile import Profile
from tracardi_dot_notation.dot_accessor import DotAccessor


class DeleteTraitAction(ActionRunner):

    def __init__(self, **kwargs):
        if 'delete' not in kwargs:
            raise ValueError("Please define delete in config section.")
        if not isinstance(kwargs['delete'], list):
            raise ValueError("Please define delete as list not {}.".format(type(kwargs['delete'])))

        self.delete = kwargs['delete']

    async def run(self, payload: dict):

        dot = DotAccessor(
            self.profile,
            self.session,
            payload if isinstance(payload, dict) else None,
            self.event,
            self.flow)

        for value in self.delete:
            del dot[value]

        profile = Profile(**dot.profile)

        self.profile.replace(profile)

        return Result(port="payload", value=payload)


def register() -> Plugin:
    return Plugin(
        start=False,
        spec=Spec(
            module='tracardi.process_engine.action.v1.traits.delete_trait_action',
            className='DeleteTraitAction',
            inputs=['payload'],
            outputs=["payload"],
            init={
                "delete": [
                    "payload@undefied",
                    "profile@undefined",
                ]
            },
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Delete Trait',
            desc='Deletes traits from profile or payload. Accepts dotted notation as definition of a filed to be '
                 'deleted. Returns payload.',
            type='flowNode',
            width=100,
            height=100,
            icon='remove',
            group=["Traits"]
        )
    )



from tracardi_plugin_sdk.domain.register import Plugin, Spec, MetaData
from tracardi_plugin_sdk.action_runner import ActionRunner
from tracardi_plugin_sdk.domain.result import Result


class InjectAction(ActionRunner):

    def __init__(self, value):
        self.value = value

    async def run(self, void):
        return Result(value=self.value, port="value")


def register() -> Plugin:
    return Plugin(
        start=True,
        debug=True,
        spec=Spec(
            module='tracardi.process_engine.action.v1.inject_action',
            className='InjectAction',
            inputs=[],
            outputs=["value"],
            init={"value": None},
            manual='inject_action',
            version='0.1',
            license="MIT",
            author="Risto Kowaczewski"
        ),
        metadata=MetaData(
            name='Inject',
            desc='Injector.',
            keywords=['start node'],
            type='flowNode',
            width=100,
            height=100,
            icon='json',
            group=["Input/Output"]
        )
    )



import asyncio
from uuid import uuid4

import pytest

from tracardi.domain.entity import Entity
from tracardi.domain.payload.tracker_payload import TrackerPayload
from tracardi.domain.profile import Profile
from tracardi.domain.session import Session


def test_suite():
    @pytest.mark.asyncio
    async def test_session_exists_profile_not_exists():
        non_existent_profile_id = str(uuid4())

        # Create Session
        session_id = str(uuid4())
        session = Session(id=session_id)
        await session.storage().save()

        # Profile does not exist
        session.profile = Profile(id=non_existent_profile_id)

        # Session exists and has profile equal to some random profile that do not exists
        tracker_payload = TrackerPayload(
            session=Entity(id=session_id),
            profile=None,
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

        assert isinstance(profile, Profile)
        assert isinstance(session, Session)

        assert profile.id != non_existent_profile_id  # profile can not be generated from non existent profile_id
        assert session.id == session_id

        await session.storage().delete()
        await profile.storage().delete()

    @pytest.mark.asyncio
    async def test_session_exists_profile_exists_conflict():
        # Create Profile for Tracker
        tracker_profile = Profile.new()
        await tracker_profile.storage().save()

        # Create Profile for Session
        session_profile = Profile.new()
        await session_profile.storage().save()

        # Create Session
        session_id = str(uuid4())
        session = Session(id=session_id)
        session.profile = session_profile
        await session.storage().save()

        tracker_payload = TrackerPayload(
            session=Entity(id=session.id),
            profile=Entity(id=tracker_profile.id),
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

        assert isinstance(profile, Profile)
        assert isinstance(session, Session)

        assert profile.id == session_profile.id  # profile id must be form session
        assert session.id == session_id

        # Remove
        await session.storage().delete()
        await tracker_profile.storage().delete()
        await session_profile.storage().delete()

    @pytest.mark.asyncio
    async def test_session_exists_profile_exists():
        # Create Profile for Session
        new_profile = Profile.new()
        await new_profile.storage().save()

        # Create Session
        session_id = str(uuid4())
        session = Session(id=session_id)
        session.profile = new_profile
        await session.storage().save()

        tracker_payload = TrackerPayload(
            session=Entity(id=session.id),
            profile=new_profile,
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

        assert isinstance(profile, Profile)
        assert isinstance(session, Session)

        assert profile.id == new_profile.id  # profile id must be form session
        assert session.id == session_id

        # Remove
        await session.storage().delete()
        await profile.storage().delete()

    @pytest.mark.asyncio
    async def test_session_not_exists_profile_exists():
        # Session does not exist
        # Profile exists

        # Create Profile for Session
        profile_id = str(uuid4())
        profile = Profile(id=profile_id)
        await profile.storage().save()

        # Create Session
        session_id = str(uuid4())
        session = Session(id=session_id)
        session.profile = profile

        tracker_payload = TrackerPayload(
            session=Entity(id=session.id),
            profile=Entity(id=profile.id),
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

        assert isinstance(profile, Profile)
        assert isinstance(session, Session)

        assert profile.id == profile_id
        assert session.id == session_id

        # Remove
        await session.storage().delete()
        await profile.storage().delete()

    @pytest.mark.asyncio
    async def test_session_not_exists_profile_not_exists():
        # Create Profile for Session
        profile_id = str(uuid4())
        profile = Profile(id=profile_id)

        # Create Session
        session_id = str(uuid4())
        session = Session(id=session_id)
        session.profile = profile

        tracker_payload = TrackerPayload(
            session=Entity(id=session.id),
            profile=Entity(id=profile.id),
            source=Entity(id="scope")
        )

        profile, session = await tracker_payload._get_profile_and_session()

        assert profile.id != profile_id  # Must generate new profile, this may be forged
        assert session.id == session_id

        # Remove
        await session.storage().delete()
        await profile.storage().delete()

    asyncio.set_event_loop(asyncio.new_event_loop())
    loop = asyncio.get_event_loop()
    loop.run_until_complete(test_session_exists_profile_not_exists())
    loop.run_until_complete(test_session_exists_profile_exists_conflict())
    loop.run_until_complete(test_session_exists_profile_exists())
    loop.run_until_complete(test_session_not_exists_profile_exists())
    loop.run_until_complete(test_session_not_exists_profile_not_exists())
    loop.close()
